/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: DatabaseImpl.java,v 1.21 2003/12/14 18:02:14 vgritsenko Exp $
 */

package org.apache.xindice.client.xmldb.embed;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.client.xmldb.CommonConfigurable;
import org.apache.xindice.core.Database;
import org.apache.xindice.server.Xindice;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.xml.dom.DOMParser;

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.ErrorCodes;
import org.xmldb.api.base.XMLDBException;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * implements XML:DB's <code>Database</code> interface providing
 * embedded access to a Xindice database.
 *
 * @author <a href="mailto:kstaken@xmldatabases.org">Kimbro Staken</a>
 * @author <a href="mailto:james.bates@amplexor.com">James Bates</a>
 * @author <a href="mailto:vladimir@apache.org">Vladimir R. Bossicard</a>
 * @version CVS $Revision: 1.21 $, $Date: 2003/12/14 18:02:14 $
 */
public class DatabaseImpl extends CommonConfigurable implements org.xmldb.api.base.Database {

    private static final Log log = LogFactory.getLog(DatabaseImpl.class);

    /**
     * Prefix used to denote XML:DB URI's that should use this driver
     */
    public static final String DRIVER_NAME = "xindice-embed";

    /**
     * XML:DB conformance level of this driver
     */
    private String CONFORMANCE_LEVEL = "0";

    /**
     * Database instance
     */
    private Database database;

    /**
     * Creates new <code>DatabaseImpl</code> instance. The configuration is
     * loaded from the file defined in the PROP_XINDICE_CONFIGURATION system
     * variable.
     */
    public DatabaseImpl() throws IOException, XindiceException {
        init();
    }

    /**
     * Create a new DatabaseImpl object with a copy of the properties
     * from the DatabaseImpl parameter.
     *
     * This allows to pass properties such as PROP_XINDICE_CONFIGURATION
     * to this instance. Usually this is done by instantiating not this
     * class, but <code>org.apache.xindice.client.xmldb.DatabaseImpl</code>,
     * set all the necessary parameters, and then get a collection.
     *
     * @param config from which the initial parameters for this
     *        DatabaseImpl object are copied.
     */
    public DatabaseImpl(CommonConfigurable config) throws IOException, XindiceException {
        super(config);
        init();
    }

    /**
     * Init this database instance.
     *
     * This is only a temporarly solution since the question of a new init
     * method is raised in the xmldb:api group. Another solution could be to
     * use the Configurable interface and only create the database when the
     * getCollection method is called.
     *
     * @throws IOException
     * @throws XindiceException
     */
    private void init() throws IOException, XindiceException {
        Configuration config = loadConfiguration();

        this.database = Database.getDatabase(config);
        if (null == this.database) {
            log.fatal("Unable to configure database");
            throw new XindiceException("Unable to configure database");
        }

        if (log.isDebugEnabled()) {
            log.info("Database name: '" + this.database.getName() + "'");
        }
    }

    protected Configuration loadConfiguration() throws IOException, XindiceException  {
        Configuration config;
        String configDir = null;

        String configFile = null;
        try {
            // Try configuration first
            configFile = getProperty(Xindice.PROP_XINDICE_CONFIGURATION);
        } catch (XMLDBException ignored) {
        }
        if (configFile == null) {
            // Fallback to system property
            configFile = System.getProperty(Xindice.PROP_XINDICE_CONFIGURATION);
        }

        if (configFile != null && !configFile.equals("")) {
            if (log.isInfoEnabled()) {
                log.info("Specified configuration file: '" + configFile + "'");
            }
            FileInputStream configXMLFile = new FileInputStream(configFile);

            config = new Configuration(DOMParser.toDocument(configXMLFile), false);
            configDir = new File(configFile).getAbsoluteFile().getParent();
        } else {
            if (log.isInfoEnabled()) {
                log.info("No configuration file specified, going with the default configuration");
            }

            config = new Configuration(DOMParser.toDocument(Xindice.DEFAULT_CONFIGURATION), false);
        }

        config = config.getChild("root-collection", false);

        String dbRoot = config.getAttribute(Database.DBROOT, Database.DBROOT_DEFAULT);
        if (!new File(dbRoot).isAbsolute()) {
            // Let's see if the property was specified.
            String home = null;
            try {
                // Try configuration first
                home = getProperty(Xindice.PROP_XINDICE_DB_HOME);
            } catch (XMLDBException ignored) {
            }
            if (home == null) {
                // Fallback to system property
                home = System.getProperty(Xindice.PROP_XINDICE_DB_HOME);
            }

            if (home != null) {
                dbRoot = new File(home + File.separator + dbRoot).getCanonicalPath();
            } else if (configDir != null) {
                dbRoot = configDir + File.separator + dbRoot;
            } else {
                log.warn("The database configuration file is not specified and there was no "
                         + Xindice.PROP_XINDICE_DB_HOME + " property set, "
                         + "so Xindice was unable to determine a database location. "
                         + "Database will be created relative to the current directory.");
                dbRoot = new File("." + File.separator + dbRoot).getCanonicalPath();
            }
            config.setAttribute(Database.DBROOT, dbRoot);
        }

        return config;
    }

    /**
     * Determines whether this <code>Database</code> implementation  can handle
     * the URI. It should return true if the Database instance knows how to
     * handle the URI and false otherwise.
     *
     * @param uri the URI to check for.
     * @return true if the URI can be handled, false otherwise.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     */
    public boolean acceptsURI(String uri) throws XMLDBException {
        return ((uri != null) && uri.startsWith(getName() + "://"));
    }

    /**
     * Retrieves a <code>Collection</code> instance based on the URI provided
     * in the <code>uri</code> parameter. The format of the URI is defined in the
     * documentation for DatabaseManager.getCollection().<p/>
     *
     * Authentication is handled via username and password however it is not
     * required that the database support authentication. Databases that do not
     * support authentication MUST ignore the
     * <code>username</code> and <code>password</code> if those provided are not
     * null.
     *
     * @param uri the URI to use to locate the collection.
     * @param password The password to use for authentication to the database or
     *    null if the database does not support authentication.
     * @return A <code>Collection</code> instance for the requested collection or
     *  null if the collection could not be found.
     * @return The <code>Collection</code> instance
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     *  <code>ErrroCodes.PERMISSION_DENIED</code> If the <code>username</code>
     *    and <code>password</code> were not accepted by the database.
     */
    public Collection getCollection(String uri, String userName, String password) throws XMLDBException {
        /* TODO: introduce authentication some day */

        if (!acceptsURI(uri)) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL: " + uri);
        }

        /* Chop off driver prefix, and '://' */
        uri = uri.substring(getName().length() + 3);

        /* Extract host name & port, if present */
        int firstSlash = uri.indexOf('/');
        if (firstSlash == -1) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL (must have '/'): " + uri);
        }

        /* Extract collection name */
        String collPath = uri.substring(firstSlash);
        if (!collPath.startsWith("/")) {
            throw new XMLDBException(ErrorCodes.INVALID_URI,
                                     "Invalid URL (collection name must start with '/'): " + uri);
        }

        // find the database name. We just skip the first slash
        int colIndex = collPath.indexOf('/', 1);

        // We assume there's no collection specified
        String dbName = collPath.substring(1);
        String colName = "/";

        // if colIndex isn't -1 then we need to pick out the db and collection
        if (colIndex != -1) {
            dbName = collPath.substring(1, colIndex);

            // The rest of the name locates the collection
            colName = collPath.substring(colIndex);
            if (colName.equals("")) {
                colName = "/";
            }
        }

        // TODO: Is this correct behavior?
        if (!database.getName().equals(dbName)) {
            throw new XMLDBException(ErrorCodes.NO_SUCH_DATABASE,
                                     "Unknown database (must be '" + database.getName() + "'): " + uri);
        }

        try {
            return new CollectionImpl(database, colName);
        } catch (XMLDBException e) {
            if (e.errorCode == ErrorCodes.NO_SUCH_COLLECTION) {
                // per getCollection contract, return null if not found
                return null;
            }
            throw e;
        }
    }

    /**
     * Returns the prefix used in XML:DB to denote URI's that this driver can
     * handle.
     *
     * @return the prefix driver name
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getName() throws XMLDBException {
        return DRIVER_NAME;
    }

    /**
     * Returns the XML:DB API Conformance level for the implementation. This can
     * be used by client programs to determine what functionality is available to
     * them.
     *
     * @return the XML:DB API conformance level for this implementation.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getConformanceLevel() throws XMLDBException {
        return CONFORMANCE_LEVEL;
    }
}
