/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: DatabaseImpl.java,v 1.17 2003/08/16 03:47:36 vgritsenko Exp $
 */

package org.apache.xindice.client.xmldb.xmlrpc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.client.xmldb.CommonConfigurable;
import org.apache.xmlrpc.XmlRpc;

import org.xmldb.api.base.Collection;
import org.xmldb.api.base.Database;
import org.xmldb.api.base.ErrorCodes;
import org.xmldb.api.base.XMLDBException;

/**
 * implements XML:DB's <code>Database</code> interface using XML-RPC to
 * communicate with the Xindice server.
 *
 * Note this class is a database <em>driver</em>, and one class of this database
 * could be used to connect to <em>many</em> different databases.
 *
 * @author <a href="mailto:james.bates@amplexor.com">James Bates</a>
 * @version CVS $Revision: 1.17 $, $Date: 2003/08/16 03:47:36 $
 */
public class DatabaseImpl extends CommonConfigurable implements Database {

    private static final Log log = LogFactory.getLog(DatabaseImpl.class);

    /**
     * Property name for the xml-rpc service location.
     */
    private static final String PROP_SERVICE_LOCATION = "service-location";

    /**
     * Property name for the SAX parser xml-rpc will use.
     */
    private static final String PROP_XMLRPC_DRIVER = "xmlrpc-driver";

    /**
     * System property name for the service location
     */
    private static final String SYSPROP_SERVICE_LOCATION = "xindice.xmlrpc.service-location";

    /**
     * System property name for the SAX parser xml-rpc will use in case
     * there were no configuration property passed
     */
    private static final String SYSPROP_XMLRPC_DRIVER = "xindice.xmlrpc.driver";

    /**
     *  Default value of the xmlrpc-driver property
     */
    private static final String DEFAULT_XMLRPC_DRIVER = "xerces";

    /**
     * Default path to the XML-RPC service in the web server
     */
    private static final String DEFAULT_SERVICE_LOCATION = "/xindice/";

    /**
     * Prefix used to denote XML:DB URI's that should use this driver
     */
    static String DRIVER_NAME = "xindice";

    /**
     * XML:DB conformance level of this driver
     */
    private String CONFORMANCE_LEVEL = "0";

    /**
     * Ensures that XML-RPC initialized just once
     */
    private static boolean xmlRpcInitialized;

    /**
     * Location of the XML-RPC service in the web server
     */
    private String serviceLocation;

    /**
     * Create a new DatabaseImpl object.
     */
    public DatabaseImpl() throws XMLDBException {
        super();
    }

    /**
     * Create a new DatabaseImpl object with a copy of the properties
     * from the DatabaseImpl parameter.
     *
     * @param config from which the initial parameters for this
     *        DatabaseImpl object are copied.
     */
    public DatabaseImpl(CommonConfigurable config) throws XMLDBException {
        super(config);
    }

    /**
     * Determines whether this <code>Database</code> implementation  can handle
     * the URI. It should return true if the Database instance knows how to
     * handle the URI and false otherwise.
     *
     * @param uri the URI to check for.
     * @return true if the URI can be handled, false otherwise.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     */
    public boolean acceptsURI(String uri) throws XMLDBException {

        return ((uri != null) && uri.startsWith(getName() + "://"));
    }

    /**
     * Initialize XML-RPC static properties: encoding, keep-alive, SAX driver.
     * Initialize XML-RPC service location.
     *
     * @throws XMLDBException if specified (or default, if none specified) SAX driver
     *         class could not be loaded
     */
    private void initialize() throws XMLDBException {
        synchronized(getClass()) {
            if (!xmlRpcInitialized) {
                XmlRpc.setEncoding("UTF8");
                XmlRpc.setKeepAlive(true);

                /*
                 * Determine the SAXparser the xmlrpc client will use.
                 * In priority order:
                 *   DatabaseImpl xmlrpc-driver property
                 *      (passed in the xmlrpcDriver parameter)
                 *   System property "xindice.xmlrpc.driver"
                 *   Default value "xerces"
                 */
                String xmlrpcDriver = getProperty(PROP_XMLRPC_DRIVER);
                if (xmlrpcDriver == null) {
                    xmlrpcDriver = System.getProperty(SYSPROP_XMLRPC_DRIVER);
                    if (xmlrpcDriver == null) {
                        xmlrpcDriver = DEFAULT_XMLRPC_DRIVER;
                    }
                }

                try {
                    if (log.isDebugEnabled()) {
                        log.debug("Using SAX Driver: '" + xmlrpcDriver + "'");
                    }
                    XmlRpc.setDriver(xmlrpcDriver);
                } catch (Exception e) {
                    throw new XMLDBException(ErrorCodes.VENDOR_ERROR,
                                             "SAX Driver " + xmlrpcDriver + " is not available", e);
                }
            }

            xmlRpcInitialized = true;
        }

        if (this.serviceLocation == null) {
            /*
             * Determine the path in the web server to the XML-RPC service.
             * In priority order:
             *   DatabaseImpl service-location property
             *      (passed in the serviceLocation parameter)
             *   System property "xindice.xmlrpc.service-location"
             *   Default value "/xindice/"
             */
            serviceLocation = getProperty(PROP_SERVICE_LOCATION);
            if (serviceLocation == null) {
                serviceLocation = System.getProperty(SYSPROP_SERVICE_LOCATION);
                if (serviceLocation == null) {
                    serviceLocation = DEFAULT_SERVICE_LOCATION;
                }
            }

            if (!serviceLocation.startsWith("/")) {
                serviceLocation = "/" + serviceLocation;
            }
            if (!serviceLocation.endsWith("/")) {
                serviceLocation = serviceLocation + "/";
            }

            if (log.isDebugEnabled()) {
                log.debug("Using Service Location: '" + serviceLocation + "'");
            }
        }
    }

    /**
     * Retrieves a <code>Collection</code> instance based on the URI provided
     * in the <code>uri</code> parameter. The format of the URI is defined in the
     * documentation for DatabaseManager.getCollection().<p/>
     *
     * Authentication is handled via username and password however it is not
     * required that the database support authentication. Databases that do not
     * support authentication MUST ignore the
     * <code>username</code> and <code>password</code> if those provided are not
     * null.
     *
     * @param uri the URI to use to locate the collection.
     * @param password The password to use for authentication to the database or
     *    null if the database does not support authentication.
     * @return A <code>Collection</code> instance for the requested collection or
     *  null if the collection could not be found.
     * @return The <code>Collection</code> instance
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     *  <code>ErrroCodes.INVALID_URI</code> If the URI is not in a valid format. <br />
     *  <code>ErrroCodes.PERMISSION_DENIED</code> If the <code>username</code>
     *    and <code>password</code> were not accepted by the database.
     */
    public Collection getCollection(String uri, String userName, String password) throws XMLDBException {
        /* TODO: introduce authentication some day */

        if (!acceptsURI(uri)) {
            throw new XMLDBException(ErrorCodes.INVALID_URI, "Invalid URL: " + uri);
        }

        /* Chop off driver prefix, and '://' */
        uri = uri.substring(getName().length() + 3);

        /* Extract host name & port, if present */
        int firstSlash = uri.indexOf('/');
        if (firstSlash == -1) {
            throw new XMLDBException(ErrorCodes.INVALID_URI, "Invalid URL (must have '/'): " + uri);
        }

        String collPath = uri.substring(firstSlash);
        String hostPort = uri.substring(0, firstSlash);

        /* Absent host defaults to localhost and standard Xindice HTTP port */
        if (hostPort.equals("")) {
            hostPort = "127.0.0.1:8888";
        }

        // Init XML-RPC in case it is not initialized yet. Initialize serviceLocation
        initialize();

        try {
            return new CollectionImpl(hostPort, serviceLocation, collPath);
        } catch (XMLDBException e) {
            if (e.errorCode == ErrorCodes.NO_SUCH_COLLECTION) {
                // per getCollection contract, return null if not found
                return null;
            }
            throw e;
        }
    }

    /**
     * Returns the prefix used in XML:DB to denote URI's that this driver can
     * handle.
     *
     * @return the prefix driver name
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getName() throws XMLDBException {
        return DRIVER_NAME;
    }

    /**
     * Returns the XML:DB API Conformance level for the implementation. This can
     * be used by client programs to determine what functionality is available to
     * them.
     *
     * @return the XML:DB API conformance level for this implementation.
     * @exception XMLDBException with expected error codes.<br />
     *  <code>ErrorCodes.VENDOR_ERROR</code> for any vendor
     *  specific errors that occur.<br />
     */
    public String getConformanceLevel() throws XMLDBException {
        return CONFORMANCE_LEVEL;
    }
}
