/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: Database.java,v 1.34 2003/12/24 15:23:08 vgritsenko Exp $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
// import org.apache.xindice.core.meta.TimeRecord;
import org.apache.xindice.core.query.QueryEngine;
import org.apache.xindice.server.Xindice;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.ConfigurationException;
import org.apache.xindice.util.Named;
import org.apache.xindice.util.XindiceException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

/**
 * Database is the primary container for the Xindice Database Engine.
 *
 * @version CVS $Revision: 1.34 $, $Date: 2003/12/24 15:23:08 $
 */
public final class Database extends Collection implements Named {

    private static final Log log = LogFactory.getLog(Database.class);

    public static final String DBROOT = "dbroot";
    public static final String NAME = "name";
    private static final String QUERYENGINE = "queryengine";
    private static final String COLKEY = "database.xml";
    private static final String DATABASE = "database";
    private static final String METADATA = "use-metadata";

    public static final String DBROOT_DEFAULT = "./db/";

    private static final Map databases = new HashMap(); // String to Database

    /**
     * This will return an instance of a Database for the given
     * name if one has already been loaded, otherwise it will
     * create a new instance.
     *
     * @param config
     * @return Database
     * @throws ConfigurationException if database name is missing in the configuration
     */
    public static Database getDatabase(Configuration config) {

        String name = config.getAttribute(Database.NAME);

        // No name in the config file ... can't get the database
        if (null == name) {
            throw new ConfigurationException("Database configuration didn't contain a database name");
        }

        Database database = (Database) databases.get(name);
        if (null == database) {
            // In case it's currently being added (only pay the sync hit on a miss)
            synchronized (databases) {
                // Was it created while we waited?
                database = (Database) databases.get(name);
                if (null == database) {
                    database = new Database();

                    try {
                        database.setConfig(config);
                    } catch (XindiceException x) {
                        // TODO: Configurable interface should use ConfigurationException instead of XindiceException.
                        throw new ConfigurationException(x);
                    }

                    databases.put(database.getName(), database);
                }
            }
        }

        return database;
    }

    /**
     * This will merely return an instance of a Database for the given
     *     name if one has already been loaded.
     *
     * @param name
     * @return Database
     */
    public static Database getDatabase(String name) {

        Database database = (Database) databases.get(name);
        if (null == database) {
            // in case it's currently being added (only pay the sync hit on a miss)
            synchronized (databases) {
                database = (Database) databases.get(name);
            }
        }

        return database;
    }

    public static String[] listDatabases() {
        return (String[]) databases.keySet().toArray(new String[0]);
    }

    //
    // Instance...
    //
    private DocumentCache docCache = new DocumentCache();
    private QueryEngine engine = new QueryEngine(this);
    private boolean metaEnabled = false;
    private boolean metaInit = false;
    private MetaSystemCollection metaSystemCollection = null;
    private boolean sysInit = false;
    private SystemCollection systemCollection = null;
    /** key = canonical name, value = timerecord */
    // private Map timestamps = new HashMap();

    /**
     * @see org.apache.xindice.core.DBObject#close()
     */
    public boolean close() throws DBException {
        flushConfig();

        synchronized (databases) {
            databases.remove(getName());
        }

        return true;
    }

    /**
     * flushConfig ensures that the Collection configuration has been
     * properly flushed to disk after a modification.
     */
    public void flushConfig() {
        try {
            Document d = getConfig().getElement().getOwnerDocument();
            systemCollection.getCollection(SystemCollection.CONFIGS).setDocument(COLKEY, d);
        } catch (Exception e) {
            log.error("Error Writing Configuration '" + getName() + "', for database " + getName(), e);
        }

        // observer
        DBObserver.getInstance().flushDatabaseConfig(this, getConfig());
    }

    /**
     * @see org.apache.xindice.core.Collection#getDatabase()
     */
    public Database getDatabase() {
        return this;
    }

    /**
     * getDocumentCache returns the Database-level Document Cache.
     *
     * @return The DocumentCache
     */
    public DocumentCache getDocumentCache() {
        return docCache;
    }

    /**
     * Return the MetaSystem collection for this database.
     *
     * It will return null if metadata is not enabled on this database.
     * @return MetaSystemCollection
     */
    public MetaSystemCollection getMetaSystemCollection() {
        return metaSystemCollection;
    }

    /**
     * getQueryEngine returns a reference to the Database's current
     * operating QueryEngine implementation.
     *
     * @return The QueryEngine instance
     */
    public QueryEngine getQueryEngine() {
        return engine;
    }

    /**
     * @see org.apache.xindice.core.Collection#getSystemCollection()
     */
    public SystemCollection getSystemCollection() {
        return systemCollection;
    }

    /**
     * TODO: Not used. Remove?
     * @param path
     * @return
    public synchronized TimeRecord getTime(String path) {
        TimeRecord rec = (TimeRecord) timestamps.get(path);
        return rec;
    }
     */

    /**
     * Return whether or not metadata is enabled on this database.
     * @return boolean
     */
    public boolean isMetaEnabled() {
        return metaEnabled;
    }

    /**
     * TODO: Not used. Remove?
     * methods for recording the times when meta data is enabled.
    public synchronized void recordTime(String path, long created, long modified) {
        TimeRecord rec = (TimeRecord) timestamps.get(path);
        if (null == rec) {
            rec = new TimeRecord(created, modified);
            timestamps.put(path, rec);
        } else {
            if (created > 0) {
                rec.setCreatedTime(created);
            }
            if (modified > 0) {
                rec.setModifiedTime(modified);
            }
        }
    }
     */

    /**
     * TODO: Not used. Remove?
     * @param path
    public synchronized void removeTime(String path) {
        timestamps.remove(path);
    }
     */

    /**
     * @see org.apache.xindice.util.Configurable#setConfig(org.apache.xindice.util.Configuration)
     */
    public void setConfig(Configuration config) throws XindiceException {
        super.setConfig(config);
        setCanonicalName('/' + getName());

        String dbroot = config.getAttribute(DBROOT);
        File dbrootDir = new File(dbroot);
        if (!dbrootDir.isAbsolute()) {
            // Here, DB root already should be absolute. XMLRPC, Embed, and Managed drivers take care of it.
            log.warn("The specified database root directory '" + dbroot + "' is relative. " +
                     "Using property " + Xindice.PROP_XINDICE_DB_HOME + " to resolve.");

            String home = System.getProperty(Xindice.PROP_XINDICE_DB_HOME);
            if (home == null) {
                log.warn("The specified database root directory '" + dbroot + "' is relative " +
                         "and there was no " + Xindice.PROP_XINDICE_DB_HOME + " property set, " +
                         "so Xindice was unable to determine a database location. " +
                         "Database will be created relative to the current working directory.");

                home = ".";
            }
            try {
                // In case home has been specified as relative path convert it to absolute path
                dbrootDir = new File(home, dbroot).getCanonicalFile();
            } catch (IOException e) {
                throw new XindiceException("Can't get canonical path", e);
            }
        }

        setCollectionRoot(dbrootDir);
        if (log.isInfoEnabled()) {
            log.info("Database points to " + dbrootDir.getAbsolutePath());
        }

        try {
            Configuration queryCfg = config.getChild(QUERYENGINE);
            if (queryCfg != null) {
                this.engine.setConfig(queryCfg);
            }
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("ignored exception", e);
            }
        }

        if (!sysInit) {
            this.systemCollection = new SystemCollection(this);

            try {
                this.systemCollection.init();
            } catch (XindiceException e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }

            super.addCollection(systemCollection);
            this.sysInit = true;
        }

        try {
            // Bootstrap from the database itself...  This is accomplished
            // by intercepting the setConfig call and using a Configuration
            // retrieved from the database instead of the standard config
            Document colDoc = systemCollection.getCollection(SystemCollection.CONFIGS).getDocument(COLKEY);
            if (colDoc == null) {
                DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
                colDoc = db.newDocument();
                Element root = colDoc.createElement(DATABASE);
                root.setAttribute(NAME, getName());
                colDoc.appendChild(root);
                systemCollection.getCollection(SystemCollection.CONFIGS).setDocument(COLKEY, colDoc);
            }

            super.setConfig(new Configuration(colDoc.getDocumentElement(), false));
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("ignored exception", e);
            }
        }

        // Register the Database with the VM
        // databases.put(getName(), this);

        // initialize the meta collection
        // but only if it's turned on in the config.
        String metaCfg = config.getAttribute(METADATA);
        if (metaCfg.equalsIgnoreCase("on")) {
            metaEnabled = true;
            if (!metaInit) {
                this.metaSystemCollection = new MetaSystemCollection(this);
                try {
                    this.metaSystemCollection.init();
                } catch (XindiceException e) {
                    log.warn("Error initializing the meta collection", e);
                }
                // should this attach the meta collection to the database?

                super.addCollection(metaSystemCollection);
                metaInit = true;
                log.info("Meta information initialized");
            }
        }

        // observer
        DBObserver.getInstance().setDatabaseConfig(this, getCollections(), config);
    }
}
