/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: DocumentCache.java,v 1.12 2003/12/25 20:04:40 vgritsenko Exp $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.xml.NodeSource;
import org.apache.xindice.xml.SymbolTable;
import org.apache.xindice.xml.dom.DBDocument;
import org.apache.xindice.xml.dom.DocumentImpl;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.Map;
import java.util.WeakHashMap;

/**
 * DocumentCache implements a simple Document caching system for
 * Collections.
 *
 * <small>
 * FIXME: Revisit cache implementation. Most probably, commons collections'
 *        ReferenceMap should be used instead of WeakHashMap.
 * </small>
 *
 * @version CVS $Revision: 1.12 $, $Date: 2003/12/25 20:04:40 $
 */
public final class DocumentCache {

    private static final Log log = LogFactory.getLog(DocumentCache.class);

    private Map table = new WeakHashMap();

    /**
     * Obtains document from cache
     *
     * @param col document collection
     * @param key document key
     * @return document from the cache or null if not present
     */
    public Document getDocument(Collection col, Key key) {
        Object v = table.get(new CacheKey(col, key));
        if (v == null) {
            return null;
        } else if (v instanceof Document) {
            return (Document) v;
        } else if (v instanceof byte[]) {
            try {
                SymbolTable s = col.getSymbols();
                NodeSource ns = new NodeSource(col, key);
                return new DocumentImpl((byte[]) v, s, ns);
            } catch (Exception e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }
        }
        return null;
    }

    /**
     * Stores document in the cache
     *
     * @param col document collection
     * @param key document key
     * @param bytes compressed document
     */
    public void putDocument(Collection col, Key key, byte[] bytes) {
        CacheKey ckey = new CacheKey(col, key);
        table.put(ckey, bytes);
    }

    /**
     * Stores document in the cache
     *
     * @param col document collection
     * @param key document key
     * @param doc document to store in the cache
     */
    public void putDocument(Collection col, Key key, Document doc) {
        CacheKey ckey = new CacheKey(col, key);
        table.put(ckey, doc);
    }

    /**
     * Remove document from the cache
     *
     * @param col document collection
     * @param key document key
     */
    public void removeDocument(Collection col, Key key) {
        table.remove(new CacheKey(col, key));
    }

    /**
     * Obtains value of the cache control processing instruction in this document
     * @param doc document to inspect for cache control processing instruction
     * @return cache control value
     */
    public static int getCacheControl(Document doc) {
        String cache = DBDocument.CACHE;
        NodeList childNodes = doc.getChildNodes();
        int size = childNodes.getLength();
        for (int i = 0; i < size; i++) {
            Node n = childNodes.item(i);
            if (n.getNodeType() == Node.PROCESSING_INSTRUCTION_NODE && n.getNodeName().equals(DBDocument.CACHE_CONTROL)) {
                cache = n.getNodeValue().trim();
                break;
            }
        }

        if (cache != null) {
            if (cache.equals(DBDocument.CACHE)) {
                return -1;
            } else if (cache.equals(DBDocument.NOCACHE)) {
                return 0;
            } else {
                return Integer.parseInt(cache);
            }
        } else {
            return -1;
        }
    }

    /**
     * CacheKey
     */
    private class CacheKey {
        private Collection col;
        private String strVal;
        private Key key;

        public CacheKey(Collection col, Key key) {
            this.col = col;
            this.key = key;
        }

        public Collection getCollection() {
            return col;
        }

        public Key getKey() {
            return key;
        }

        public String toString() {
            if (strVal == null) {
                strVal = col.getCanonicalDocumentName(key);
            }
            return strVal;
        }

        public int hashCode() {
            if (strVal == null) {
                strVal = col.getCanonicalDocumentName(key);
            }
            return strVal.hashCode();
        }

        public boolean equals(Object o) {
            return o instanceof CacheKey && col == ((CacheKey) o).col && key.equals(((CacheKey) o).key);
        }
    }
}
