/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: HashFiler.java,v 1.19 2003/12/22 01:38:42 vgritsenko Exp $
 */

package org.apache.xindice.core.filer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.DBException;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;
import org.apache.xindice.util.Configurable;
import org.apache.xindice.util.Configuration;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

/**
 * HashFiler is a Filer implementation based on the Paged class.  By
 * extending Paged, HashFiler inherits the ability to maintain Record
 * metadata such as creation and modification time.  It also provides
 * quite a bit more flexibility in its ability to retreive blocks of
 * data and allocate Record space.
 *
 * @deprecated This class has been temporarily (maybe indefinitely)
 *             deprecated by BTreeFiler.
 * @version CVS $Revision: 1.19 $, $Date: 2003/12/22 01:38:42 $
 */
public final class HashFiler extends Paged implements Configurable, Filer {

    private static final Log log = LogFactory.getLog(HashFiler.class);

    protected static final byte RECORD = 1;

    private static final String PAGECOUNT = "pagecount";
    private static final String PAGESIZE = "pagesize";
    private static final String MAXKEYSIZE = "maxkeysize";
    private Configuration config;
    private HashFileHeader fileHeader;


    public HashFiler() {
        super();
        fileHeader = (HashFileHeader) getFileHeader();
    }

    public void setConfig(Configuration config) {
        this.config = config;
    }

    public Configuration getConfig() {
        return config;
    }

    public void setLocation(File root, String location) {
        setFile(new File(root, location + ".tbl"));
    }

    public String getName() {
        return "HashFiler";
    }

    public boolean open() throws DBException {
        if (super.open()) {
            // These are the only properties that can be changed after creation
            fileHeader.setMaxKeySize(config.getShortAttribute(MAXKEYSIZE, fileHeader.getMaxKeySize()));
            return true;
        } else {
            return false;
        }
    }

    public boolean create() throws DBException {
        fileHeader.setPageCount(config.getLongAttribute(PAGECOUNT, fileHeader.getPageCount()));
        fileHeader.setPageSize(config.getIntAttribute(PAGESIZE, fileHeader.getPageSize()));
        fileHeader.setMaxKeySize(config.getShortAttribute(MAXKEYSIZE, fileHeader.getMaxKeySize()));
        return super.create();
    }

    private Page seekRecordPage(Key key) throws IOException {
        int hash = key.getHash();
        long pageNum = hash % fileHeader.getPageCount();
        Page p = getPage(pageNum);
        synchronized (p) {
            HashPageHeader ph = null;
            while (true) {
                ph = (HashPageHeader) p.getPageHeader();
                if (ph.getStatus() == RECORD && ph.getKeyHash() == key.getHash() && p.getKey().equals(key)) {
                    return p;
                }

                pageNum = ph.getNextCollision();
                if (pageNum == -1) {
                    return null;
                }
                p = getPage(pageNum);
            }
        }
    }

    public Record readRecord(Key key) throws DBException {
        if (key == null || key.equals("")) {
            return null;
        }
        checkOpened();
        try {
            Page startPage = seekRecordPage(key);
            if (startPage != null) {
                Value v = readValue(startPage);
                HashPageHeader sph = (HashPageHeader) startPage.getPageHeader();

                HashMap meta = new HashMap(2);
                meta.put(Record.CREATED, new Long(sph.getCreated()));
                meta.put(Record.MODIFIED, new Long(sph.getModified()));

                return new Record(key, v, meta);
            }
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("ignored exception", e);
            }
        }
        return null;
    }

    private Page seekInsertionPage(Key key) throws IOException {
        int hash = key.getHash();
        long pageNum = hash % fileHeader.getPageCount();

        // Synchronize by chain head page
        Page p = getPage(pageNum);
        synchronized (p) {
            HashPageHeader ph;
            while (true) {
                ph = (HashPageHeader) p.getPageHeader();
                if (ph.getStatus() == UNUSED || ph.getStatus() == DELETED
                        || (ph.getStatus() == RECORD && ph.getKeyHash() == key.getHash() && p.getKey().equals(key))) {
                    // Found free page
                    break;
                }

                pageNum = ph.getNextCollision();
                if (pageNum == -1) {
                    // Reached end of chain, add new page
                    Page np = getFreePage();

                    ph.setNextCollision(np.getPageNum().longValue());
                    p.write();

                    p = np;
                    ph = (HashPageHeader) p.getPageHeader();
                    ph.setNextCollision(NO_PAGE);
                    break;
                }

                // Go to next page in chain
                p = getPage(pageNum);
            }

            // Here we have a page
            long t = System.currentTimeMillis();
            if (ph.getStatus() == UNUSED || ph.getStatus() == DELETED) {
                // This is a new Record
                fileHeader.incRecordCount();
                ph.setCreated(t);
            }
            ph.setModified(t);
            ph.setStatus(RECORD);

            // Write modifications to the page header before existing synchronization block
            // This will prevent other threads from getting this same page
            p.write();
        }

        return p;
    }

    public boolean writeRecord(Key key, Value value) throws DBException {
        // Check that key is not larger than space on the page
        if (key == null || key.equals("") || key.getLength() > fileHeader.getPageSize() - fileHeader.getPageHeaderSize()) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid key: '" + key + "'");
        }
        if (value == null) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Invalid null value");
        }
        checkOpened();
        Page p = null;
        try {
            p = seekInsertionPage(key);
            p.setKey(key);
            writeValue(p, value);
            p = null;
        } catch (Exception e) {
            throw new FilerException(FaultCodes.DBE_CANNOT_CREATE, "Exception: " + e);
        } finally {
            if (p != null) {
                p.getPageHeader().setStatus(DELETED);
                try {
                    p.write();
                } catch (IOException ignored) {
                    // Double exception
                }
            }
        }

        flush();
        return true;
    }

    public boolean deleteRecord(Key key) throws DBException {
        if (key == null || key.equals("")) {
            return false;
        }
        checkOpened();
        try {
            int hash = key.getHash();
            long pageNum = hash % fileHeader.getPageCount();

            Page page = getPage(pageNum);
            synchronized(page) {
                HashPageHeader prevHead = null;
                HashPageHeader pageHead = null;

                Page prev = null;
                while (true) {
                    pageHead = (HashPageHeader) page.getPageHeader();
                    if (pageHead.getStatus() == RECORD && pageHead.getKeyHash() == key.getHash() && page.getKey().equals(key)) {
                        break;
                    }

                    pageNum = pageHead.getNextCollision();
                    if (pageNum == NO_PAGE) {
                        return false;
                    }
                    prev = page;
                    prevHead = pageHead;
                    page = getPage(pageNum);
                }

                if (prev != null) {
                    prevHead.setNextCollision(pageHead.nextCollision);
                    pageHead.setNextCollision(NO_PAGE);
                    prev.write();
                }

                unlinkPages(page);
            }

            fileHeader.decRecordCount();
            flush();

            return true;
        } catch (Exception e) {
            if (log.isWarnEnabled()) {
                log.warn("ignored exception", e);
            }
        }
        return false;
    }

    public long getRecordCount() throws DBException {
        checkOpened();
        return fileHeader.getRecordCount();
    }

    public RecordSet getRecordSet() throws DBException {
        checkOpened();
        return new HashFilerRecordSet();
    }

    public void flush() throws DBException {
        super.flush();
    }

    /**
     * HashFilerRecordSet that does not use a BTree.
     */
    private class HashFilerRecordSet implements RecordSet {
        private List keys = new ArrayList();
        private Iterator enum;

        public HashFilerRecordSet() {
            try {
                Page p = null;
                HashPageHeader ph = null;
                long pageNum = 0;

                while (pageNum < fileHeader.getTotalCount()) {
                    p = getPage(pageNum);
                    ph = (HashPageHeader) p.getPageHeader();
                    if (ph.getStatus() == RECORD) {
                        keys.add(p.getKey());
                    }

                    pageNum++;
                }

                enum = keys.iterator();
            } catch (Exception e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }
        }

        public synchronized Key getNextKey() {
            return (Key) enum.next();
        }

        public synchronized Record getNextRecord() throws DBException {
            return readRecord((Key) enum.next());
        }

        public synchronized Value getNextValue() throws DBException {
            return getNextRecord().getValue();
        }

        public synchronized boolean hasMoreRecords() {
            return enum.hasNext();
        }
    }

    ////////////////////////////////////////////////////////////////////

    public FileHeader createFileHeader() {
        return new HashFileHeader();
    }

    public FileHeader createFileHeader(boolean read) throws IOException {
        return new HashFileHeader(read);
    }

    public FileHeader createFileHeader(long pageCount) {
        return new HashFileHeader(pageCount);
    }

    public FileHeader createFileHeader(long pageCount, int pageSize) {
        return new HashFileHeader(pageCount, pageSize);
    }

    public PageHeader createPageHeader() {
        return new HashPageHeader();
    }

    /**
     * HashFileHeader
     */
    private final class HashFileHeader extends FileHeader {
        private long totalBytes = 0;

        public HashFileHeader() {
        }

        public HashFileHeader(long pageCount) {
            super(pageCount);
        }

        public HashFileHeader(long pageCount, int pageSize) {
            super(pageCount, pageSize);
        }

        public HashFileHeader(boolean read) throws IOException {
            super(read);
        }

        public synchronized void read(RandomAccessFile raf) throws IOException {
            super.read(raf);
            totalBytes = raf.readLong();
        }

        public synchronized void write(RandomAccessFile raf) throws IOException {
            super.write(raf);
            raf.writeLong(totalBytes);
        }

        /** The total number of bytes in use by the file */
        public synchronized void setTotalBytes(long totalBytes) {
            this.totalBytes = totalBytes;
            setDirty();
        }

        /** The total number of bytes in use by the file */
        public synchronized long getTotalBytes() {
            return totalBytes;
        }
    }

    /**
     * HashPageHeader
     */
    private final class HashPageHeader extends PageHeader {
        private long created = 0;
        private long modified = 0;
        private long nextCollision = NO_PAGE;

        public HashPageHeader() {
        }

        public HashPageHeader(DataInputStream dis) throws IOException {
            super(dis);
        }

        public synchronized void read(DataInputStream dis) throws IOException {
            super.read(dis);

            if (getStatus() == UNUSED) {
                return;
            }

            created = dis.readLong();
            modified = dis.readLong();
            nextCollision = dis.readLong();
        }

        public synchronized void write(DataOutputStream dos) throws IOException {
            super.write(dos);
            dos.writeLong(created);
            dos.writeLong(modified);
            dos.writeLong(nextCollision);
        }

        public synchronized void setRecordLen(int recordLen) {
            fileHeader.setTotalBytes((fileHeader.totalBytes - getRecordLen()) + recordLen);
            super.setRecordLen(recordLen);
        }

        /** UNIX-time when this record was created */
        public synchronized void setCreated(long created) {
            this.created = created;
            setDirty();
        }

        /** UNIX-time when this record was created */
        public synchronized long getCreated() {
            return created;
        }

        /** UNIX-time when this record was last modified */
        public synchronized void setModified(long modified) {
            this.modified = modified;
            setDirty();
        }

        /** UNIX-time when this record was last modified */
        public synchronized long getModified() {
            return modified;
        }

        /** The next page for a Record collision (if any) */
        public synchronized void setNextCollision(long nextCollision) {
            this.nextCollision = nextCollision;
            setDirty();
        }

        /** The next page for a Record collision (if any) */
        public synchronized long getNextCollision() {
            return nextCollision;
        }
    }
}
