/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: InlineMetaService.java,v 1.6 2003/08/13 03:18:45 vgritsenko Exp $
 */

package org.apache.xindice.core.meta.inline;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.FaultCodes;
import org.apache.xindice.core.data.Value;

/**
 * If the documents in a collection have headers, the Collection
 * object holds an instance of this class.  All services the collection
 * needs for working with inline metadata are provided by this class.
 *
 * @version CVS $Revision: 1.6 $, $Date: 2003/08/13 03:18:45 $
 */
public class InlineMetaService {

    private static Log log = LogFactory.getLog(InlineMetaService.class);

    /**
     * The known readers.  All readers for header versions up to
     * and including the current version must be present in this array.
     */
    public static final InlineMetaReader[] readerByVersion = {new NullReader(), new ResourceTypeReader()};

    /**
     * The writer corresponding to the current header version.
     */
    public static final InlineMetaWriter currentWriter = new ResourceTypeWriter();

    /**
     * The InlineMetaMap class associated with the current header
     * version.  When the Collection fills in an InlineMetaMap object
     * prior to a store, it is working with this class.
     */
    public static final Class currentMapClass = ResourceTypeReader.ResourceTypeMap.class;

    /**
     * Get an <code>InlineMetaMap</code> instance corresponding to the
     * current header version.  Used by <code>Collection</code> to inform
     * the InlineMeta code of the metadata values for the entry being stored.
     *
     * @return InlineMetaMap instance
     */
    public InlineMetaMap getEmptyMap() {
        try {
            return (InlineMetaMap) currentMapClass.newInstance();
        } catch (Exception e) {
            String msg = "unable to create instance of current inline metadata map class '" + currentMapClass.getName() + "'";
            /*
             * This can only happen as a result of a programming error,
             * so log it and throw a runtime.  Shouldn't happen in production.
             */
            if (log.isFatalEnabled()) {
                log.fatal(msg, e);
            }
            throw new IllegalStateException(msg);
        }
    }

    /**
     * Create a Value object containing the appropriate header and
     * data.  This utility method takes the metadata and the data
     * for the entry to be stored in the database and creates a
     * <code>Value</code> object containing both, ready to store.
     *
     * @param map containing metadata
     * @param data containing entry to be stored in database
     * @param offset in the data to the beginning of the entry
     * @param length of the entry
     * @return Value object containing the header + entry, ready to store
     * @throws InlineMetaException if the metadata is incomplete
     */
    public Value createValue(InlineMetaMap map, byte[] data, int offset, int length) throws InlineMetaException {

        byte[] metadata = currentWriter.getMetadata(map);

        return InlineHeaderBuilder.createValue(currentWriter.getVersion(), metadata, data, offset, length);
    }

    /**
     * Returns true if a reader is available for the specified
     * version.
     *
     * @param version of the header
     * @return true if a reader is available for the given version
     */
    public boolean haveReaderForVersion(int version) {
        return version < readerByVersion.length;
    }

    /**
     * Given a <code>Value</code> object from the database,
     * disassemble it into its component metadata and entry data
     * components.
     *
     * @param rawValue containing raw header + entry data from the backing store
     * @return DatabaseEntry containing metadata in an <code>InlineMetaMap</code>
     *         and entry data in a <code>Value</code>.
     * @throws InlineMetaException if missing the reader for the header, or if
     *         the header is corrupted.
     */
    public DatabaseEntry readDatabaseEntry(Value rawValue) throws InlineMetaException {
        if (log.isDebugEnabled()) {
            log.debug("readDatabaseEntry: rawValue: length=" + rawValue.getLength());
        }

        byte[] rawData = rawValue.getData();

        if (log.isDebugEnabled()) {
            log.debug("readDatabaseEntry: rawData: length=" + rawData.length + " byte 0: " + rawData[0] + " byte 1: " + rawData[1]);
        }

        /*
         * Read the header.
         */

        int version = rawData[1];
        if (!haveReaderForVersion(version)) {
            throw new InlineMetaException(FaultCodes.GEN_CRITICAL_ERROR, "No inline metadata reader available for version " + version);
        }
        InlineMetaReader reader = readerByVersion[version];
        if (log.isDebugEnabled()) {
            log.debug("readHeader: reader=" + reader);
        }

        InlineMetaMap map = reader.read(rawData, 2, rawData[0] - 2);
        if (log.isDebugEnabled()) {
            log.debug("readDatabaseEntry: map: type=" + map.get("type"));
        }

        /*
         * Exract the data into a Value object.
         */

        Value value = new Value(rawData, rawData[0], rawData.length - rawData[0]);
        if (log.isDebugEnabled()) {
            log.debug("readDatabaseEntry: value: type=" + map.get("type"));
        }

        return new DatabaseEntry(map, value);
    }

    /**
     * Utility class for returning an InlineMetaMap and Value
     * from <code>readDatabaseEntry(Value)</code>.
     */
    public static class DatabaseEntry {
        public final InlineMetaMap map;
        public final Value value;

        public DatabaseEntry(InlineMetaMap map, Value value) {
            this.map = map;
            this.value = value;
        }
    }
}
