/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * CVS $Id: ManagedServer.java,v 1.6 2003/08/19 21:08:27 vgritsenko Exp $
 */

package org.apache.xindice.server;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.Database;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.XindiceException;
import org.apache.xindice.xml.dom.DOMParser;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * Creates an registers a database instance for use of the managed driver. This class
 * may be used to create the database the managed driver uses (as is the case in the test
 * cases) but it is not required. Any method may be used to create the database instance
 * that the managed driver talks to. As long as the instance name exists when the driver
 * fires up all will be good.
 *
 * @author <a href="mailto:kevin@rocketred.com.au">Kevin O'Neill</a>
 * @version CVS $Revision: 1.6 $, $Date: 2003/08/19 21:08:27 $
 */
public class ManagedServer {

    private static final Log log = LogFactory.getLog(ManagedServer.class);

    private Database db = null;
    private boolean running = false;


    public synchronized void start() throws Exception {
        if (!running) {
            configure();
            running = true;
        } else {
            log.warn("Start called on a running server, ignored");
        }
    }

    public void stop() throws Exception {
        if (running) {
            db.close();
            db = null;
        } else {
            log.warn("Stop called on a stopped server, ignored");
        }
    }

    public void configure() throws IOException, XindiceException {
        db = Database.getDatabase(loadConfiguration());

        if (null == db) {
            log.fatal("Unable to configure database");
            throw new XindiceException("Unable to configure database");
        } else if (log.isInfoEnabled()) {
            log.info("Database name: '" + db.getName() + "'");
        }
    }

    protected Configuration loadConfiguration() throws IOException, XindiceException {
        Configuration config;
        String configDir = null;
        String configFile = System.getProperty(Xindice.PROP_XINDICE_CONFIGURATION);
        if (configFile != null && !configFile.equals("")) {
            if (log.isInfoEnabled()) {
                log.info("Specified configuration file: '" + configFile + "'");
            }
            FileInputStream configXMLFile = new FileInputStream(new File(configFile));

            config = new Configuration(DOMParser.toDocument(configXMLFile), false);
            configDir = new File(configFile).getAbsoluteFile().getParent();
        } else {
            if (log.isInfoEnabled()) {
                log.info("No configuration file specified, going with the default configuration");
            }

            config = new Configuration(DOMParser.toDocument(Xindice.DEFAULT_CONFIGURATION), false);
        }

        config = config.getChild("root-collection", false);

        String dbRoot = config.getAttribute(Database.DBROOT, Database.DBROOT_DEFAULT);
        if (!new File(dbRoot).isAbsolute()) {
            // Let's see if the property was specified.
            String home = System.getProperty(Xindice.PROP_XINDICE_DB_HOME);
            if (home != null) {
                dbRoot = new File(home + File.separator + dbRoot).getCanonicalPath();
            } else if (configDir != null) {
                dbRoot = configDir + File.separator + dbRoot;
            } else {
                dbRoot = new File("." + File.separator + dbRoot).getCanonicalPath();
                log.warn("The database configuration file is not specified and there was no "
                         + Xindice.PROP_XINDICE_DB_HOME + " property set, "
                         + "so Xindice was unable to determine a database location. "
                         + "Database will be created relative to the current directory.");
            }
            config.setAttribute(Database.DBROOT, dbRoot);
        }

        return config;
    }
}
